/*
 * packetmodeller/logger.c
 *
 * Copyright (C) 2008-2009 Reto Gantenbein
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <arpa/inet.h>
#include <net/ethernet.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "typedef.h"
#include "logger.h"

/** Print raw data in hex to stdout */
void
dump_data(char *data, int len)
{
    int i;

    printf("(%d bytes) data: ", len);
    for(i = 1; i <= len; i++)
    {
        printf("%02hhx ", *(data + (i-1)));
        if (i % 16 == 0)
            printf("\n                 ");
    }
    printf("\n");
}

/** Print message header struct */
void
dump_header(struct msg_header *header)
{
    int i;

    printf("(%i bytes) type: ", sizeof(header->type));
    switch (header->type)
    {
        case NODE_MSG:
            printf("Node_Msg");
            break;
        case MODEL_MSG:
            printf("Model_Msg");
            break;
        case REGISTER_MSG:
            printf("Register_Msg");
            break;
        case ACK_MSG:
            printf("Acknowledge_Msg");
            break;
        default:
            printf("unknown message");
            break;
    }
    printf(" (%i)\n", header->type);
    printf("(%i bytes) id: %s\n", sizeof(header->id), header->id);
}

/** Print information from register message */
void
dump_register_info(struct register_msg *msg)
{
    printf("(%i bytes) msg_id: %i\n", sizeof(msg->msg_id), msg->msg_id);
    printf("(%i bytes) name: %s\n", sizeof(msg->name), msg->name);
    printf("(%i bytes) ip: %s\n", sizeof(msg->ip), inet_ntoa(msg->ip));
    printf("(%i bytes) port: %i\n", sizeof(msg->port), ntohs(msg->port));
    printf("(%i bytes) virt_mac: %s\n", sizeof(msg->virt_mac), msg->virt_mac);
}

/** Print information about wlan parameters sent to model */
void
dump_wlan_info(struct node_msg *msg)
{
    printf("(%i bytes) channel: %i\n", sizeof(msg->channel), msg->channel);
    printf("(%i bytes) transmit_power: %i\n", \
        sizeof(msg->transmit_power), msg->transmit_power);
    printf("(%i bytes) pkt_length: %i\n", \
        sizeof(msg->pkt_length), msg->pkt_length);
}

/** Print IP header and return packet length */
u_int16_t
dump_ip_header(struct iphdr *ip_hdr)
{
    //structs and types defined in "netinet/ip.h"
    printf("(%i bytes) header: ", sizeof(struct iphdr));
    printf("tos 0x%x, ", ip_hdr->tos);
    printf("ttl %i, ", ip_hdr->ttl);
    printf("id %i, " , ntohs(ip_hdr->id));
    printf("proto ");
    switch (ip_hdr->protocol)
    {
        case 1:
            printf("ICMP ");
            break;
        case 6:
            printf("TCP ");
            break;
        case 17:
            printf("UDP ");
            break;
        default:
            //for more, see: /etc/protocols
            printf("unknown ");
            break;
    }
    printf("(%i), ", ip_hdr->protocol);
    u_int16_t packet_length = ntohs(ip_hdr->tot_len);
    printf("length %i\n", packet_length);
    printf("                   ");
    u_char ip[INET_ADDRSTRLEN];
    inet_ntop(AF_INET, (struct in_addr *) &(ip_hdr->saddr), \
        ip, INET_ADDRSTRLEN);
    printf("saddr %s, ", ip);
    inet_ntop(AF_INET, (struct in_addr *) &(ip_hdr->daddr), \
        ip, INET_ADDRSTRLEN);
    printf("daddr %s\n", ip);

    return packet_length;
}

/** Print TCP packet */
void
dump_tcp_packet(char *tcp_packet, u_int16_t pkt_size)
{
    // structs and types defined in "netinet/tcp.h"
    struct tcphdr *tcp_hdr = (struct tcphdr *) tcp_packet;

    printf("(%i bytes) tcp packet:\n", pkt_size);
    printf("(%i bytes) header: ", sizeof(struct tcphdr));
    printf("source %i, ", ntohs(tcp_hdr->source));
    printf("dest %i\n", ntohs(tcp_hdr->dest));
    dump_data(tcp_packet + sizeof(struct tcphdr), \
        pkt_size - sizeof(struct tcphdr));
}

/** Print IP packet */
void
dump_ip_packet(char* ip_packet, int pkt_size)
{
    struct iphdr *ip_hdr = (struct iphdr *) ip_packet;

    printf("(%i bytes) ip packet:\n", pkt_size);
    if (pkt_size != dump_ip_header(ip_hdr))
    {
        printf("WARNING: ");
        printf("defined packet size doesn't match with received packet size!\n");
    }
    u_int8_t proto = ip_hdr->protocol;
    switch (proto)
    {
        case 6:
            dump_tcp_packet(ip_packet + sizeof(struct iphdr), \
                pkt_size - sizeof(struct iphdr));
            break;
        default:
            dump_data(ip_packet + sizeof(struct iphdr), \
                pkt_size - sizeof(struct iphdr));
            break;
    }
}

/** Print Ethernet header */
u_int16_t
dump_ethernet_header(char* ehdr)
{
    // structs and types defined in "net/ethernet.h"
    int i;
    struct ether_header *ether_hdr = (struct ether_header *) ehdr;

    printf("(%i bytes) header: ether_dhost ", sizeof(struct ether_header));
    for (i = 0; i < ETH_ALEN; i++)
    {
        printf("%02hhx", (*ether_hdr).ether_dhost[i]);
        if (i < (ETH_ALEN - 1))
            printf(":");
    }
    printf("\n                   ether_shost ");
    for (i = 0; i < ETH_ALEN; i++)
    {
        printf("%02hhx", (*ether_hdr).ether_shost[i]);
        if (i < (ETH_ALEN - 1))
            printf(":");
    }
    printf("\n                   ether_type ");
    u_int16_t ether_type = ntohs((*ether_hdr).ether_type);
    if (ether_type == ETHERTYPE_PUP)
        printf("Xerox PUP\n");
    if (ether_type == ETHERTYPE_IP)
        printf("IP\n");
    if (ether_type == ETHERTYPE_ARP)
        printf("ARP\n");
    if (ether_type == ETHERTYPE_REVARP)
        printf("Reverse ARP\n");
    return ether_type;
}

/** Print PacketModeller <-> WlanModel message */
void
dump_msg(const char *pkt, int frame_length)
{
    char *wlan_packet;

    printf("****************************************\n");
    printf("dump_msg():\n");
    dump_header((struct msg_header *) pkt);
    if (isNodeMessage((struct msg_header *) pkt))
        dump_wlan_info((struct node_msg *) pkt);
    if (isRegisterMessage((struct msg_header *) pkt))
        dump_register_info((struct register_msg *) pkt);
    if (isNodeMessage((struct msg_header *) pkt))
    {
        wlan_packet = (char *) &(*(struct node_msg *) pkt).wlan_pkt;
//      dump_data(wlan_packet, frame_length);
    }
    if (isModelMessage((struct msg_header *) pkt))
    {
        wlan_packet = (char *) &(*(struct model_msg *) pkt).wlan_pkt;
//      dump_data(wlan_packet, frame_length);
    }
    if (isModelMessage((struct msg_header *) pkt) || \
        isNodeMessage((struct msg_header *) pkt))
    {
        printf("(%i bytes) ethernet packet:\n", frame_length);
        u_int16_t ether_header_size = sizeof(struct ether_header);
        u_int16_t payload_length = frame_length - ether_header_size;
        u_int16_t type = dump_ethernet_header(wlan_packet);
        switch (type)
        {
            case ETHERTYPE_IP:
                dump_ip_packet(wlan_packet + ether_header_size, payload_length);
                break;
            default:
                printf("****************************************\n");
                dump_data(wlan_packet + ether_header_size, payload_length);
                break;
        }
    }
    printf("****************************************\n");
}

/** Do nothing but imitate the printf() signature */
int
ignore_dbg(const char* text, ...)
{
}
